/*
    HPYLM.cpp
    $Id: hpylm.cpp,v 1.5 2010/03/17 16:41:57 daichi Exp $

*/
#include "hpylm.h"
#include "ngram.h"
#include "util.h"
#include "debug.h"
#include "random.h"
#include <cstdio>
#include <algorithm>

using namespace std;

HPYLM::HPYLM (int order_)
{
	tree = new ngram (NULL, EOS, this);
	order = order_;
	lexicon = 0;
	d.resize (order + 1);
	theta.resize (order + 1);
	fill (d.begin(), d.end(), 0.5);
	fill (theta.begin(), theta.end(), 0.1);
}

HPYLM::~HPYLM ()
{
	delete tree;
}

bool
HPYLM::add_data (witerator w)
{
	words.push_back (w);
}

bool
HPYLM::add_customer (witerator w)
{
	return _add_customer (order, w, w - 1, tree);
}

bool
HPYLM::_add_customer (int n, witerator word, witerator history, ngram *context)
{
	ngram *child;

	if (n == 0)
		return context->increment_customer (*word);
	else {
		if ((child = find_node (context->children, *history)) == NULL)
		{
			/* no descendant found */
			if ((child = new ngram(context, *history)) == NULL)
			{
				error("add_customer: can't create ngram.\n");
				return false;
			}
			insert_node (context->children, child);
		}
		return _add_customer (n - 1, word, history - 1, child);
	}
}

bool
HPYLM::remove_customer (witerator w)
{
	return _remove_customer (order, w, w - 1, tree);
}

bool
HPYLM::_remove_customer (int n, witerator word, witerator history, ngram *context)
{
	ngram *child;

	if (n == 0)
		return context->decrement_customer (*word);
	else {
		/* find child */
		if ((child = find_node (context->children, *history)) == NULL)
		{
			error("remove_customer: can't find the context for %d.\n", *history);
			return false;
		} else {
			return _remove_customer (n - 1, word, history - 1, child);
		}
	}
}

/*
 *  inference
 *
 */

bool
HPYLM::gibbs (int iter)
{
	int j, n;
	vector<witerator>::iterator it;

	for (j = 0; j < iter; j++)
	{
		printf("Gibbs iteration %d..\n", j + 1);
		
		random_shuffle (words.begin(), words.end(), irand);
		
		for (it = words.begin(); it != words.end(); it++)
		{
			remove_customer (*it);
			add_customer (*it);
		}
		for (n = 0; n <= order; n++)
		{
			d[n] = hpy_d_posterior (n);
			theta[n] = hpy_theta_posterior (n);
			printf("d[%d] = %g, theta[%d] = %g\n",
			       n, d[n], n, theta[n]);
		}
	}
}

bool
HPYLM::hyper_estimate ()
{
	int n;

	for (n = 0; n <= order; n++)
	{
		d[n] = hpy_d_posterior (n);
		theta[n] = hpy_theta_posterior (n);
	}
}
	

/*
 *  probability
 *
 */

double
HPYLM::hpy_probability (int *word, int *history)
{
	return hpy_probability ((witerator)word, (witerator)history);
}

double
HPYLM::hpy_probability (witerator word, witerator history)
{
	return _hpy_probability (order, word, history, tree);
}

double
HPYLM::_hpy_probability (int n, witerator word, witerator history, ngram *context)
{
	ngram *child;

	if (n == 0)
		return context->hpy_probability (*word);
	else {
		/* find child */
		if ((child = find_node (context->children, *history)) == NULL)
			/* not found */
			return context->hpy_probability (*word);
		else
			/* found */
			return _hpy_probability (n - 1, word, history - 1, child);
	}
}

/*
 *  misc
 *
 */

double
hpylm_customers (ngram *node, void *arg)
{
	hashmap *children = node->children;
	
	if (children->size() == 0)
		return (double) node->ncounts;
	else
		return (double) node->ncounts
			+ hash_fold (children, (foldfunc)hpylm_customers, NULL);
}

int
HPYLM::customers ()
{
	return (int) hash_fold (tree->children, (foldfunc)hpylm_customers, NULL);
}

void
HPYLM::print ()
{
	tree->print ();
}

ngram *
HPYLM::find_context (witerator history)
{
	return _find_context (order, history, tree);
}

ngram *
HPYLM::_find_context (int n, witerator history, ngram *context)
{
	ngram *child;
	
	if (n == 0)
		return context;
	else {
		/* find child */
		if ((child = find_node (context->children, *history)) == NULL)
			/* not found */
			return context;
		else
			return _find_context (n - 1, history - 1, child);
	}
}

