/*
    random.cpp
    $Id: random.cpp,v 1.2 2010/03/17 16:41:57 daichi Exp $

*/
#include <cstdio>
#include <cstdlib>
#include <cmath>
#include "mt.h"
#include "random.h"
#include "util.h"
#include "debug.h"

/*
 *  main
 *
 */

void
fisher_yates_shuffle (int *v, int n)
{
	int i, j, t;

	for (i = n - 1; i >= 0; i--)
	{
		j = (int)(RANDOM * (i + 1));
		t = v[i];
		v[i] = v[j];
		v[j] = t;
	}
}

int
bernoulli (double p)	/* return 1 with probability p */
{
	double r = RANDOM;
	if (r < p)
		return 1;
	else
		return 0;
}

int
draw_discrete (double *table, int n)
{
	int i;
	double r, s;

	for (i = 0, s = 0; i < n; i++)
		s += table[i];
	r = RANDOM * s;
	for (i = 0, s = 0; i < n; i++)
	{
		s += table[i];
		if (r <= s)
			return i;
	}
	/* for rounding error */
	dprintf("draw_discrete: r = %lf, s = %lf\n", r, s);
	return n - 1;
}

int
draw_discrete_int (int *table, int n)
{
	int i;
	double r, s;

	for (i = 0, s = 0; i < n; i++)
		s += table[i];
	r = RANDOM * s;
	for (i = 0, s = 0; i < n; i++)
	{
		s += table[i];
		if (r <= s)
			return i;
	}
	/* for rounding error */
	dprintf("draw_discrete_int: r = %lf, s = %lf\n", r, s);
	return n - 1;
}

int
draw_probability (double *table, int n)
{
	int i;
	double s = 0, r = RANDOM;

	for (i = 0; i < n; i++)
	{
		s += table[i];
		if (r <= s)
			return i;
	}
	/* for rounding error */
	dprintf("draw_probability: r = %lf, s = %lf\n", r, s);
	return n - 1;
}

int
draw_likelihood (double *lik, int n)
{
	int i;
	double z;
	static int maxlen = 0;
	static double *table = NULL;

	if (n > maxlen)
	{
		if ((table = (double *)realloc(table, n * sizeof(double))) == NULL)
		{
			perror("realloc");
			return -1;
		}
		maxlen = n;
	}
	
	for (i = 0, z = 0; i < n; i++)
		z = logsumexp(z, lik[i], (i == 0));
	for (i = 0; i < n; i++)
		table[i] = exp(lik[i] - z);

	return draw_probability (table, n);
}


double
sgamrand (double a)
{
	double x, y, z;
	double u, v, w, b, c, e;
	int accept = 0;
	if (a <= 1)
	{
		/* a <= 1. Johnk's generator. Devroye (1986) p.418 */
		e = exprand();
		do {
			x = pow(RANDOM, 1 / a);
			y = pow(RANDOM, 1 / (1 - a));
		} while (x + y > 1);
		return (e * x / (x + y));
	} else {
		/* a > 1. Best's rejection algorithm. Devroye (1986) p.410 */
		b = a - 1;
		c = 3 * a - 0.75;
		while (accept != 1)
		{
			/* generate */
			u = RANDOM;
			v = RANDOM;
			w = u * (1 - u);
			y = sqrt(c / w) * (u - 0.5);
			x = b + y;
			if (x >= 0)
			{
				z = 64 * w * w * w * v * v;
				if (z <= 1 - (2 * y * y) / x)
				{
					accept = 1;
				} else {
					if (log(z) <= 2 * (b * log(x / b) - y))
						accept = 1;
				}
			}
		}
		return x;
	}
}

double
gamrand (double a, double b)
{
	return sgamrand(a) / b;
}

double
exprand (void)
{
	return (- log(RANDOM));
}

double
betarand (double a, double b)
{
	double x = sgamrand(a), y = sgamrand(b);
	return x / (x + y);
}

int
irand (int n)
{
	return int(RANDOM * n);
}

#if 0
int
main (int argc, char *argv[])
{
	int i, k, n;
	double z, buf[4];

	buf[0] = 0.1;
	buf[1] = 0.2;
	buf[2] = 0.3;
	buf[3] = 0.4;
	for (i = 0, z = 0; i < 4; i++)
		z += buf[i];

	n = atoi(argv[1]);

	for (i = 0; i < n; i++)
	{
		k = draw_discrete(buf, 4, z);
		printf("k = %d\n", k);
	}

}
#endif

